import re


_EMPTY_LINE = re.compile(r"^\s*$")
_SPACE = re.compile(r"\s+")

def is_empty(text: str) -> bool:
  return bool(_EMPTY_LINE.match(text))

def clean_spaces(text: str) -> str:
  return _SPACE.sub(" ", text.strip())


from typing import Iterator, Generator
from resource_segmentation import split, Resource, Segment

from ..llm import LLM
from .types import Fragment
from .chunk import ChunkRange


def split_into_chunks(llm: LLM, fragments_iter: Iterator[Fragment], max_chunk_tokens_count: int):
  for index, group in enumerate(split(
    resources=_gen_resources(llm, fragments_iter),
    max_segment_count=max_chunk_tokens_count,
    gap_rate=0.15,
    tail_rate=0.5,
  )):
    head_index: int
    tail_index: int
    fragments_count: int
    body_index, body_end_index, body_tokens_count = _range_of_group_part(group.body)

    if group.head:
      head_index, head_end_index, _ = _range_of_group_part(group.head)
      assert head_end_index + 1 == body_index, "Head must be continuous with body"
    else:
      head_index = body_index

    if group.tail:
      tail_index, tail_end_index, _ = _range_of_group_part(group.tail)
      fragments_count = tail_end_index - head_index + 1
      assert body_end_index + 1 == tail_index, "Body must be continuous with tail"
    else:
      tail_index = body_end_index + 1
      fragments_count = tail_index - head_index

    yield ChunkRange(
      index=index,
      head_remain_tokens=group.head_remain_count,
      tail_remain_tokens=group.tail_remain_count,
      head_index=head_index,
      body_index=body_index,
      tail_index=tail_index,
      fragments_count=fragments_count,
      tokens_count=body_tokens_count,
    )

def _gen_resources(llm: LLM, fragments_iter: Iterator[Fragment]) -> Generator[Resource[int], None, None]:
  for index, fragment in enumerate(fragments_iter):
    yield Resource(
      count=llm.count_tokens_count(fragment.text),
      start_incision=fragment.start_incision,
      end_incision=fragment.end_incision,
      payload=index,
    )

def _range_of_group_part(target: list[Resource[int] | Segment[int]]) -> tuple[int, int]:
  start_index: int | None = None
  previous_index: int = 0
  tokens_count: int = 0
  for resource in _iter_group_part(target):
    index = resource.payload
    if start_index is None:
      start_index = index
    else:
      assert index == previous_index + 1, "Resources in group part must be continuous"
    previous_index = index
    tokens_count += resource.count

  assert start_index is not None, "Group part must contain at least one resource"
  return start_index, previous_index, tokens_count

def _iter_group_part(target: list[Resource[int] | Segment[int]]) -> Generator[Resource[int], None, None]:
  for item in target:
    if isinstance(item, Resource):
      yield item
    elif isinstance(item, Segment):
      for resource in item.resources:
        yield resource


from typing import Generator, TypeGuard
from enum import auto, Enum
from xml.etree.ElementTree import Element


class TextPosition(Enum):
  WHOLE_DOM = auto()
  TEXT = auto()
  TAIL = auto()

# element, position, parent
TextDescription = tuple[Element, TextPosition, Element | None]

_IGNORE_TAGS = (
  "title", "link", "style", "css", "img", "script", "metadata",
  "{http://www.w3.org/1998/Math/MathML}math", # TODO: 公式是正文，也要读进去，暂时忽略避免扰乱得了。
)

_TEXT_LEAF_TAGS = (
  "a", "b", "br", "hr", "span", "em", "strong", "label",
)

def search_texts(element: Element, parent: Element | None = None) -> Generator[TextDescription, None, None]:
  if element.tag in _IGNORE_TAGS:
    return

  if any(c.tag not in _TEXT_LEAF_TAGS for c in element):
    if _is_not_empty_str(element.text):
      yield element, TextPosition.TEXT, parent
    for child in element:
      if child.tag in _TEXT_LEAF_TAGS:
        yield child, TextPosition.WHOLE_DOM, element
      else:
        yield from search_texts(child, element)
      if _is_not_empty_str(child.tail):
        yield child, TextPosition.TAIL, element
  else:
    yield element, TextPosition.WHOLE_DOM, parent

def _is_not_empty_str(text: str | None) -> TypeGuard[str]:
  if text is None:
    return False
  for char in text:
    if char not in (" ", "\n"):
      return True
  return False



from os import PathLike
from pathlib import Path
from tempfile import mkdtemp
from shutil import rmtree
from resource_segmentation import Incision

from .llm import LLM
from .epub import HTMLFile
from .zip_context import ZipContext
from .translation import translate as _translate, Fragment, Language, ProgressReporter


def translate(
      llm: LLM,
      source_path: PathLike,
      translated_path: PathLike,
      target_language: Language,
      user_prompt: str | None = None,
      working_path: PathLike | None = None,
      max_chunk_tokens_count: int = 3000,
      max_threads_count: int = 1,
      report_progress: ProgressReporter | None = None,
    ) -> None:

  source_path = Path(source_path)
  translated_path = Path(translated_path)
  working_path = Path(working_path) if working_path else None
  report_progress = report_progress or (lambda _: None)

  _Translator(
    llm=llm,
    target_language=target_language,
    user_prompt=user_prompt,
    max_chunk_tokens_count=max_chunk_tokens_count,
    max_threads_count=max_threads_count,
    report_progress=report_progress,
  ).do(
    source_path=source_path,
    translated_path=translated_path,
    working_path=working_path,
  )

class _Translator:
  def __init__(
        self,
        llm: LLM,
        target_language: Language,
        user_prompt: str | None,
        max_chunk_tokens_count: int,
        max_threads_count: int,
        report_progress: ProgressReporter,
      ) -> None:

    self._llm: LLM = llm
    self._target_language: Language = target_language
    self._user_prompt: str | None = user_prompt
    self._max_chunk_tokens_count: int = max_chunk_tokens_count
    self._max_threads_count: int = max_threads_count
    self._report_progress: ProgressReporter = report_progress

  def do(self, source_path: Path, translated_path: Path, working_path: Path | None) -> None:
    is_temp_workspace = not bool(working_path)
    working_path = working_path or Path(mkdtemp())
    try:
      temp_dir = _clean_path(working_path / "temp")
      temp_dir.mkdir(parents=True, exist_ok=True)
      cache_path = working_path / "cache"

      context = ZipContext(
        epub_path=Path(source_path),
        temp_dir=temp_dir,
      )
      context.replace_ncx(lambda texts: self._translate_ncx(
        texts=texts,
        cache_path=cache_path,
        report_progress=lambda p: self._report_progress(p * 0.1)),
      )
      self._translate_spine(
        context=context,
        cache_path=cache_path,
        report_progress=lambda p: self._report_progress(0.1 + p * 0.8),
      )
      context.archive(translated_path)
      self._report_progress(1.0)

    finally:
      if is_temp_workspace:
        rmtree(working_path, ignore_errors=True)

  def _translate_ncx(self, texts: list[str], cache_path: Path, report_progress: ProgressReporter) -> list[str]:
    return list(_translate(
      llm=self._llm,
      cache_path=cache_path,
      max_chunk_tokens_count=self._max_chunk_tokens_count,
      max_threads_count=1,
      target_language=self._target_language,
      user_prompt=self._user_prompt,
      report_progress=report_progress,
      gen_fragments_iter=lambda: (
        Fragment(
          text=text,
          start_incision=Incision.IMPOSSIBLE,
          end_incision=Incision.IMPOSSIBLE,
        )
        for text in texts
      ),
    ))

  def _translate_spine(self, context: ZipContext, cache_path: Path, report_progress: ProgressReporter):
    spine_paths_iter = iter(list(context.search_spine_paths()))
    spine: tuple[Path, HTMLFile] | None = None
    translated_texts: list[str] = []
    translated_count: int = 0

    for translated_text in _translate(
      llm=self._llm,
      gen_fragments_iter=lambda: _gen_fragments(context),
      cache_path=cache_path,
      max_chunk_tokens_count=self._max_chunk_tokens_count,
      max_threads_count=self._max_threads_count,
      target_language=self._target_language,
      user_prompt=self._user_prompt,
      report_progress=report_progress,
    ):
      did_touch_end = False

      if spine and translated_count >= len(translated_texts):
        spine_path, spine_file = spine
        spine_file.write_texts(translated_texts)
        context.write_spine_file(spine_path, spine_file)
        spine = None

      while not spine:
        spine_path = next(spine_paths_iter, None)
        if spine_path is None:
          spine = None
          did_touch_end = True
          break
        spine_file = context.read_spine_file(spine_path)
        if spine_file.texts_length == 0:
          continue
        spine = (spine_path, spine_file)
        translated_texts = [""] * spine_file.texts_length
        translated_count = 0
        break

      translated_texts[translated_count] = translated_text
      translated_count += 1

      if did_touch_end:
        break

    if spine:
      spine_path, spine_file = spine
      if translated_count > 0:
        spine_file.write_texts(translated_texts)
      context.write_spine_file(spine_path, spine_file)

def _gen_fragments(context: ZipContext):
  for spine_path in context.search_spine_paths():
    spine_file = context.read_spine_file(spine_path)
    for text in spine_file.read_texts():
      yield Fragment(
        text=text,
        start_incision=Incision.IMPOSSIBLE,
        end_incision=Incision.IMPOSSIBLE,
      )

def _clean_path(path: Path) -> Path:
  if path.exists():
    if path.is_file():
      path.unlink()
    elif path.is_dir():
      rmtree(path, ignore_errors=True)
  return path


from enum import Enum
from dataclasses import dataclass
from resource_segmentation import Incision


@dataclass
class Fragment:
  text: str
  start_incision: Incision
  end_incision: Incision

class Language(Enum):
  SIMPLIFIED_CHINESE = "zh-Hans"
  TRADITIONAL_CHINESE = "zh-Hant"
  ENGLISH = "en"
  FRENCH = "fr"
  GERMAN = "de"
  SPANISH = "es"
  RUSSIAN = "ru"
  ITALIAN = "it"
  PORTUGUESE = "pt"
  JAPANESE = "ja"
  KOREAN = "ko"

_LANGUAGE_NAMES = {
  Language.SIMPLIFIED_CHINESE: "简体中文",
  Language.TRADITIONAL_CHINESE: "繁体中文",
  Language.ENGLISH: "英语",
  Language.FRENCH: "法语",
  Language.GERMAN: "德语",
  Language.SPANISH: "西班牙语",
  Language.RUSSIAN: "俄语",
  Language.ITALIAN: "意大利语",
  Language.PORTUGUESE: "葡萄牙语",
  Language.JAPANESE: "日语",
  Language.KOREAN: "韩语",
}

def language_chinese_name(language: Language) -> str:
  return _LANGUAGE_NAMES[language]


import re

# HTML 规定了一系列自闭标签，这些标签需要改成非自闭的，因为 EPub 格式不支持
# https://www.tutorialspoint.com/which-html-tags-are-self-closing
_EMPTY_TAGS = (
  "br",
  "hr",
  "input",
  "col",
  "base",
  "meta",
  "area",
)

_EMPTY_TAG_PATTERN = re.compile(
  r"<(" + "|".join(_EMPTY_TAGS) + r")(\s[^>]*?)\s*/?>"
)

def to_html(content: str) -> str:
  return re.sub(_EMPTY_TAG_PATTERN, lambda m: f"<{m.group(1)}{m.group(2)}>", content)

def to_xml(content: str) -> str:
  return re.sub(_EMPTY_TAG_PATTERN, lambda m: f"<{m.group(1)}{m.group(2)} />", content)


from math import ceil
from typing import Callable, Iterator, Generator
from pathlib import Path
from concurrent.futures import as_completed, ThreadPoolExecutor
from xml.etree.ElementTree import Element

from ..llm import LLM
from ..xml import encode_friendly

from .types import language_chinese_name, Fragment, Language
from .store import Store
from .splitter import split_into_chunks
from .chunk import match_fragments, Chunk
from .utils import is_empty, clean_spaces


ProgressReporter = Callable[[float], None]

def translate(
      llm: LLM,
      gen_fragments_iter: Callable[[], Iterator[Fragment]],
      cache_path: Path | None,
      target_language: Language,
      user_prompt: str | None,
      max_chunk_tokens_count: int,
      max_threads_count: int,
      report_progress: ProgressReporter,
    ) -> Generator[str, None, None]:

  if user_prompt is not None:
    user_prompt = _normalize_user_input(user_prompt.splitlines())

  store = Store(cache_path) if cache_path else None
  chunk_ranges = list(split_into_chunks(
    llm=llm,
    fragments_iter=gen_fragments_iter(),
    max_chunk_tokens_count=max_chunk_tokens_count,
  ))
  with ThreadPoolExecutor(max_workers=max_threads_count) as executor:
    futures = [
      executor.submit(lambda chunk=chunk: (chunk, _translate_chunk(
        llm=llm,
        store=store,
        chunk=chunk,
        target_language=target_language,
        user_prompt=user_prompt,
      )))
      for chunk in match_fragments(
        llm=llm,
        target_language=target_language,
        chunk_ranges_iter=iter(chunk_ranges),
        fragments_iter=gen_fragments_iter(),
      )
    ]
    def _generate_chunks_from_futures():
      try:
        for future in as_completed(futures):
          yield future.result()
      except Exception as err:
        for future in futures:
          if not future.done():
            future.cancel()
        raise err

    yield from _sort_translated_texts_by_chunk(
      target=_generate_chunks_from_futures(),
      total_tokens_count=sum(chunk.tokens_count for chunk in chunk_ranges),
      report_progress=report_progress,
    )

def _sort_translated_texts_by_chunk(
      target: Iterator[tuple[Chunk, list[str]]],
      total_tokens_count: int,
      report_progress: ProgressReporter,
    ) -> Iterator[list[str]]:

  buffer: list[tuple[Chunk, list[str]]] = []
  wanna_next_index: int = 0
  translated_tokens_count: int = 0

  for chunk, translated_texts in target:
    buffer.append((chunk, translated_texts))
    if wanna_next_index == chunk.index:
      buffer.sort(key=lambda e: e[0].index)
      to_clear: list[list[str]] = []

      for chunk, translated_texts in buffer:
        if chunk.index > wanna_next_index:
          break
        to_clear.append(translated_texts)
        if chunk.index == wanna_next_index:
          wanna_next_index += 1

      if to_clear:
        buffer = buffer[len(to_clear):]
        for translated_texts in to_clear:
          yield from translated_texts

    translated_tokens_count += chunk.tokens_count
    report_progress(float(translated_tokens_count) / total_tokens_count)

def _translate_chunk(
      llm: LLM,
      store: Store,
      chunk: Chunk,
      target_language: Language,
      user_prompt: str | None,
    ) -> list[str]:

    translated_texts: list[str] | None = None
    source_texts = chunk.head + chunk.body + chunk.tail
    if store is not None:
      translated_texts = store.get(chunk.hash)
      if translated_texts is not None and \
         len(source_texts) != len(translated_texts):
        translated_texts = None
        print(f"Warning: Mismatched lengths in cached translation for chunk: {chunk.hash.hex()}",)

    if translated_texts is None:
      translated_texts = [
        clean_spaces(text)
        for text in _translate_texts(
          llm=llm,
          texts=source_texts,
          texts_tokens=chunk.tokens_count,
          target_language=target_language,
          user_prompt=user_prompt,
        )
      ]
      if store is not None:
        store.put(chunk.hash, translated_texts)

    head_length = len(chunk.head)
    translated_texts = translated_texts[head_length:head_length + len(chunk.body)]

    return translated_texts

_PLAIN_TEXT_SCALE = 2.0
_XML_TEXT_SCALE = 2.5

def _translate_texts(
      llm: LLM,
      texts: list[str],
      texts_tokens: int,
      target_language: Language,
      user_prompt: str | None,
    ) -> list[str]:

  original_text = _normalize_user_input(texts)
  if original_text is None:
    return [""] * len(texts)

  user_data = original_text
  if user_prompt is not None:
    user_data = f"<rules>{user_prompt}</rules>\n\n{original_text}"

  translated_text = llm.request_text(
    template_name="translate",
    text_tag="TXT",
    user_data=user_data,
    parser=lambda r: r,
    max_tokens=ceil(texts_tokens * _PLAIN_TEXT_SCALE),
    params={
      "target_language": language_chinese_name(target_language),
      "user_prompt": user_prompt,
    },
  )
  request_element = Element("request")

  for i, fragment in enumerate(texts):
    fragment_element = Element("fragment", attrib={
      "id": str(i + 1),
    })
    fragment_element.text = clean_spaces(fragment)
    request_element.append(fragment_element)

  request_element_text = encode_friendly(request_element)
  request_text = f"```XML\n{request_element_text}\n```\n\n{translated_text}"

  return llm.request_xml(
    template_name="format",
    user_data=request_text,
    max_tokens=ceil(texts_tokens * _XML_TEXT_SCALE),
    parser=lambda r: _parse_translated_response(r, len(texts)),
    params={
      "target_language": language_chinese_name(target_language),
    },
  )

def _parse_translated_response(resp_element: Element, sources_count: int) -> list[str]:
  fragments: list[str | None] = [None] * sources_count
  for fragment_element in resp_element:
    if fragment_element.text is None:
      continue
    id = fragment_element.get("id", None)
    if id is None:
      continue
    index = int(id) - 1
    if index < 0 or index >= len(fragments):
      raise ValueError(f"invalid fragment id: {id}")
    fragments[index] = fragment_element.text.strip()

  # 有时 LLM 会将多段融合在一起，这里尽可能让译文靠后，将空白段留在前面。
  # 这样看起来一大段的译文对应若干小段原文，观感更好。
  for i in range(len(fragments)):
    fragment = fragments[i]
    if fragment is not None and i < len(fragments) - 1:
      next_fragment = fragments[i + 1]
      if next_fragment is None:
        fragments[i] = None
        fragments[i + 1] = fragment

  return [f or "" for f in fragments]

def _normalize_user_input(user_lines: list[str]) -> str | None:
  empty_lines_count: int = 0
  lines: list[str] = []
  for line in user_lines:
    if is_empty(line):
      empty_lines_count += 1
    else:
      if lines:
        if empty_lines_count >= 2:
          lines.append("")
          lines.append("")
        elif empty_lines_count == 1:
          lines.append("")
      lines.append(clean_spaces(line))
  if not lines:
    return None
  return "\n".join(lines)


import re

from typing import Iterable
from xml.etree.ElementTree import fromstring, tostring, Element
from .dom_operator import read_texts, append_texts
from .empty_tags import to_xml, to_html


_FILE_HEAD_PATTERN = re.compile(r"^<\?xml.*?\?>[\s]*<!DOCTYPE.*?>")
_XMLNS_IN_TAG = re.compile(r"\{[^}]+\}")
_BRACES = re.compile(r"(\{|\})")

class HTMLFile:
  def __init__(self, file_content: str):
    match = re.match(_FILE_HEAD_PATTERN, file_content)
    xml_content = re.sub(_FILE_HEAD_PATTERN, "", to_xml(file_content))
    self._head: str = match.group() if match else None
    self._root: Element = fromstring(xml_content)
    self._xmlns: str | None = self._extract_xmlns(self._root)
    self._texts_length: int | None = None

  def _extract_xmlns(self, root: Element) -> str | None:
    root_xmlns: str | None = None
    for i, element in enumerate(_all_elements(root)):
      need_clean_xmlns = True
      match = re.match(_XMLNS_IN_TAG, element.tag)

      if match:
        xmlns = re.sub(_BRACES, "", match.group())
        if i == 0:
          root_xmlns = xmlns
        elif root_xmlns != xmlns:
          need_clean_xmlns = False
      if need_clean_xmlns:
        element.tag = re.sub(_XMLNS_IN_TAG, "", element.tag)

    return root_xmlns

  def read_texts(self) -> list[str]:
    texts = list(read_texts(self._root))
    self._texts_length = len(texts)
    return texts

  def write_texts(self, texts: Iterable[str]):
    append_texts(self._root, texts)

  @property
  def texts_length(self) -> int:
    if self._texts_length is None:
      self._texts_length = 0
      for _ in read_texts(self._root):
        self._texts_length += 1
    return self._texts_length

  @property
  def file_content(self) -> str:
    file_content: str
    if self._xmlns is None:
      file_content = tostring(self._root, encoding="unicode")
      file_content = to_html(file_content)
    else:
      root = Element(
        self._root.tag,
        attrib={**self._root.attrib, "xmlns": self._xmlns},
      )
      root.extend(self._root)
      # XHTML disable <tag/> (we need replace them with <tag></tag>)
      for element in _all_elements(root):
        if element.text is None:
          element.text = ""
      file_content = tostring(root, encoding="unicode")

    if self._head is not None:
      file_content = self._head + file_content
    return file_content

def _all_elements(parent: Element):
  yield parent
  for child in parent:
    yield from _all_elements(child)


import zipfile

from typing import Generator, Callable
from pathlib import Path
from lxml.etree import parse
from .epub import EpubContent, HTMLFile


class ZipContext:
  def __init__(self, epub_path: Path, temp_dir: Path):
    with zipfile.ZipFile(epub_path, "r") as zip_ref:
      for member in zip_ref.namelist():
        target_path = temp_dir / member
        if member.endswith("/"):
          target_path.mkdir(parents=True, exist_ok=True)
        else:
          target_path.parent.mkdir(parents=True, exist_ok=True)
          with zip_ref.open(member) as source:
            with open(target_path, "wb") as file:
              file.write(source.read())

    self._temp_dir: Path = temp_dir
    self._epub_content: EpubContent = EpubContent(temp_dir)

  def archive(self, saved_path: Path):
    with zipfile.ZipFile(saved_path, "w") as zip_file:
      for file_path in self._temp_dir.rglob("*"):
        if not file_path.is_file():
          continue
        relative_path = file_path.relative_to(self._temp_dir)
        zip_file.write(
          filename=file_path,
          arcname=str(relative_path),
        )

  def search_spine_paths(self) -> Generator[Path, None, None]:
    for spine in self._epub_content.spines:
      if spine.media_type == "application/xhtml+xml":
        yield Path(spine.path)

  def read_spine_file(self, spine_path: Path) -> HTMLFile:
    with open(spine_path, "r", encoding="utf-8") as file:
      return HTMLFile(file.read())

  def write_spine_file(self, spine_path: Path, file: HTMLFile):
    with open(spine_path, "w", encoding="utf-8") as f:
      f.write(file.file_content)

  def replace_ncx(self, replace: Callable[[list[str]], list[str]]):
    ncx_path = self._epub_content.ncx_path
    if ncx_path is None:
      return

    tree = parse(ncx_path)
    root = tree.getroot()
    namespaces={ "ns": root.nsmap.get(None) }
    text_doms = []
    text_list = []

    for text_dom in root.xpath("//ns:text", namespaces=namespaces):
      text_doms.append(text_dom)
      text_list.append(text_dom.text or "")

    for index, text in enumerate(replace(text_list)):
      text_dom = text_doms[index]
      text_dom.text = self._link_translated(text_dom.text, text)

    tree.write(ncx_path, pretty_print=True)

  def _link_translated(self, origin: str, target: str) -> str:
    if origin == target:
      return origin
    else:
      return f"{origin} - {target}"


from dataclasses import dataclass
from typing import Iterator, Iterable, Generator
from hashlib import sha512
from ..llm import LLM
from .types import Fragment, Language


@dataclass
class Chunk:
  index: int
  hash: bytes
  head: list[str]
  body: list[str]
  tail: list[str]
  tokens_count: int

@dataclass
class ChunkRange:
  index: int
  head_remain_tokens: int
  tail_remain_tokens: int
  head_index: int
  body_index: int
  tail_index: int
  fragments_count: int
  tokens_count: int

  def match(self, index: int) -> bool:
    return self.head_index <= index < self.head_index + self.fragments_count

def match_fragments(
        llm: LLM,
        target_language: Language,
        chunk_ranges_iter: Iterator[ChunkRange],
        fragments_iter: Iterator[Fragment],
      ) -> Generator[Chunk, None, None]:

  for range, texts in _match_range_and_texts(
    chunk_range_iter=chunk_ranges_iter,
    fragments_iter=fragments_iter,
  ):
    head_length = range.body_index - range.head_index
    body_length = range.tail_index - range.body_index
    head = texts[:head_length]
    body = texts[head_length:head_length + body_length]
    tail = texts[head_length + body_length:]

    hash = _hash_texts_list(target_language, (head, body, tail))
    head = _crop_extra_texts(llm, head, True, range.head_remain_tokens)
    tail = _crop_extra_texts(llm, tail, False, range.tail_remain_tokens)

    yield Chunk(
      hash=hash,
      head=head,
      body=body,
      tail=tail,
      index=range.index,
      tokens_count=range.tokens_count,
    )

def _match_range_and_texts(
      chunk_range_iter: Iterator[ChunkRange],
      fragments_iter: Iterator[Fragment],
    ) -> Generator[tuple[ChunkRange, list[str]], None, None]:

  next_chunk_range: ChunkRange | None = None
  matched_chunk_ranges: list[tuple[ChunkRange, list[str]]] = []

  for index, fragment in enumerate(fragments_iter):
    while True:
      if next_chunk_range is None:
        next_chunk_range = next(chunk_range_iter, None)
        if next_chunk_range is None:
          break
      if not next_chunk_range.match(index):
        break
      matched_chunk_ranges.append((next_chunk_range, []))
      next_chunk_range = None

    if matched_chunk_ranges:
      next_matched_chunks: list[tuple[ChunkRange, list[str]]] = []
      for chunk_range, texts in matched_chunk_ranges:
        if chunk_range.match(index):
          texts.append(fragment.text)
          next_matched_chunks.append((chunk_range, texts))
        else:
          yield chunk_range, texts
      matched_chunk_ranges = next_matched_chunks

  yield from matched_chunk_ranges

def _hash_texts_list(target_language: Language, texts_iterable: Iterable[list[str]]) -> bytes:
  m = sha512()
  m.update(target_language.value.encode("utf-8"))
  for texts in texts_iterable:
    for text in texts:
      m.update(b"\x00")
      m.update(text.encode("utf-8"))
  return m.digest()

def _crop_extra_texts(llm: LLM, texts: list[str], crop_left: bool, remain_tokens_count: int):
  tokens_list: list[list[int]] = [llm.encode_tokens(text) for text in texts]
  remain_texts: list[str] = []

  for tokens in (reversed(tokens_list) if crop_left else tokens_list):
    tokens_count = len(tokens)
    if remain_tokens_count >= tokens_count:
      remain_tokens_count -= tokens_count
      remain_texts.append(llm.decode_tokens(tokens))
      if remain_tokens_count == 0:
        break
    else:
      remain_tokens = tokens[-remain_tokens_count:] if crop_left else tokens[:remain_tokens_count]
      remain_texts.append(llm.decode_tokens(remain_tokens))

  if crop_left:
    remain_texts.reverse()
  return remain_texts


from shutil import rmtree
from pathlib import Path
from typing import Iterator


class Store:
  def __init__(self, directory: Path):
    self._directory = directory

  def get(self, chunk_hash: bytes) -> list[str] | None:
    file_path = self._file_path(chunk_hash)
    if not file_path.exists() or not file_path.is_file():
      return None
    with file_path.open("r", encoding="utf-8") as file:
      return file.read().split("\n")

  def put(self, chunk_hash: bytes, lines_iter: Iterator[str]):
    file_path = self._file_path(chunk_hash)
    if file_path.exists():
      if file_path.is_file():
        file_path.unlink()
      else:
        rmtree(file_path)

    file_path.parent.mkdir(parents=True, exist_ok=True)
    with file_path.open("w", encoding="utf-8") as file:
      is_first_line = True
      for line in lines_iter:
        if is_first_line:
          is_first_line = False
        else:
          file.write("\n")
        file.write(line)

  def _file_path(self, chunk_hash: bytes) -> Path:
    return self._directory / f"{chunk_hash.hex()}.chunk"


import os
import re

from lxml.etree import parse, Element, QName
from html import escape


# TODO replace with XML
class Spine:
  def __init__(self, folder_path, base_path, item):
    self._folder_path = folder_path
    self._base_path = base_path
    self.href = item.get("href")
    self.media_type = item.get("media-type")

  @property
  def path(self) -> str:
    path = os.path.join(self._base_path, self.href)
    path = os.path.abspath(path)

    if os.path.exists(path):
      return path

    path = os.path.join(self._folder_path, self.href)
    path = os.path.abspath(path)
    return path

class EpubContent:
  def __init__(self, path: str):
    self.folder_path = path
    self._content_path = self._find_content_path(path)
    self._tree = parse(self._content_path)
    self._namespaces = { "ns": self._tree.getroot().nsmap.get(None) }
    self._spine = self._tree.xpath("//ns:spine", namespaces=self._namespaces)[0]
    self._metadata = self._tree.xpath("//ns:metadata", namespaces=self._namespaces)[0]
    self._manifest = self._tree.xpath("//ns:manifest", namespaces=self._namespaces)[0]

  def save(self):
    self._tree.write(self._content_path, pretty_print=True)

  def _find_content_path(self, path: str) -> str:
    root = parse(os.path.join(path, "META-INF", "container.xml")).getroot()
    rootfile = root.xpath(
      "//ns:container/ns:rootfiles/ns:rootfile",
      namespaces={ "ns": root.nsmap.get(None) },
    )[0]
    full_path = rootfile.attrib["full-path"]
    joined_path = os.path.join(path, full_path)

    return os.path.abspath(joined_path)

  @property
  def ncx_path(self):
    ncx_dom = self._manifest.find(".//*[@id=\"ncx\"]")
    if ncx_dom is not None:
      href_path = ncx_dom.get("href")
      base_path = os.path.dirname(self._content_path)
      path = os.path.join(base_path, href_path)
      path = os.path.abspath(path)

      if os.path.exists(path):
        return path

      path = os.path.join(self.folder_path, path)
      path = os.path.abspath(path)
      return path

  @property
  def spines(self) -> list[Spine]:
    idref_dict = {}
    index = 0

    for child in self._spine.iterchildren():
      id = child.get("idref")
      idref_dict[id] = index
      index += 1

    items = [None for _ in range(index)]
    spines = []

    for child in self._manifest.iterchildren():
      id = child.get("id")
      if id in idref_dict:
        index = idref_dict[id]
        items[index] = child

    base_path = os.path.dirname(self._content_path)

    for item in items:
      if item is not None:
        spines.append(Spine(
          folder_path=self.folder_path,
          base_path=base_path,
          item=item,
        ))

    return spines

  @property
  def title(self):
    title_dom = self._get_title()
    if title_dom is None:
      return None
    return title_dom.text

  @title.setter
  def title(self, title: str):
    title_dom = self._get_title()
    if title_dom is not None:
      title_dom.text = _escape_ascii(title)

  def _get_title(self):
    titles = self._metadata.xpath(
      "./dc:title",
      namespaces={
        "dc": self._metadata.nsmap.get("dc"),
      },
    )
    if len(titles) == 0:
      return None
    return titles[0]

  @property
  def authors(self) -> list[str]:
    return list(map(lambda x: x.text, self._get_creators()))

  @authors.setter
  def authors(self, authors):
    creator_doms = self._get_creators()
    if len(creator_doms) == 0:
      return
    parent_dom = creator_doms[0].getparent()
    index_at_parent = parent_dom.index(creator_doms[0])
    ns={
      "dc": self._metadata.nsmap.get("dc"),
      "opf": self._metadata.nsmap.get("opf"),
    }
    for author in reversed(authors):
      creator_dom = Element(QName(ns["dc"], "creator"))
      creator_dom.set(QName(ns["opf"], "file-as"), author)
      creator_dom.set(QName(ns["opf"], "role"), "aut")
      creator_dom.text = _escape_ascii(author)
      parent_dom.insert(index_at_parent, creator_dom)

    for creator_dom in creator_doms:
      parent_dom.remove(creator_dom)

  def _get_creators(self):
    return self._metadata.xpath(
      "./dc:creator",
      namespaces={
        "dc": self._metadata.nsmap.get("dc"),
      },
    )

def _escape_ascii(content: str) -> str:
  content = escape(content)
  content = re.sub(
    r"\\u([\da-fA-F]{4})",
    lambda x: chr(int(x.group(1), 16)), content,
  )
  return content